@@ -182,8 +182,10 @@ struct supplicant_task {
 	char *netpath;
 	gboolean created;
 	enum supplicant_state state;
-	gboolean noscan;
+	gboolean scanning;
 	GSList *scan_results;
+	DBusPendingCall *scan_call;
+	DBusPendingCall *result_call;
 	struct iw_range *range;
 	gboolean disconnecting;
 };
@@ -229,6 +231,66 @@ static struct supplicant_task *find_task_by_path(const char *path)
 	return NULL;
 }
 
+static int get_range(struct supplicant_task *task)
+{
+	struct iwreq wrq;
+	int fd, err;
+
+	fd = socket(PF_INET, SOCK_DGRAM, 0);
+	if (fd < 0)
+		return -1;
+
+	memset(&wrq, 0, sizeof(struct iwreq));
+	strncpy(wrq.ifr_name, task->ifname, IFNAMSIZ);
+	wrq.u.data.pointer = task->range;
+	wrq.u.data.length = sizeof(struct iw_range);
+
+	err = ioctl(fd, SIOCGIWRANGE, &wrq);
+
+	close(fd);
+
+	return err;
+}
+
+static int get_bssid(struct connman_device *device,
+				unsigned char *bssid, unsigned int *bssid_len)
+{
+	struct iwreq wrq;
+	char *ifname;
+	int ifindex;
+	int fd, err;
+
+	ifindex = connman_device_get_index(device);
+	if (ifindex < 0)
+		return -EINVAL;
+
+	ifname = connman_inet_ifname(ifindex);
+	if (ifname == NULL)
+		return -EINVAL;
+
+	fd = socket(PF_INET, SOCK_DGRAM, 0);
+	if (fd < 0) {
+		g_free(ifname);
+		return -EINVAL;
+	}
+
+	memset(&wrq, 0, sizeof(wrq));
+	strncpy(wrq.ifr_name, ifname, IFNAMSIZ);
+
+	err = ioctl(fd, SIOCGIWAP, &wrq);
+
+	g_free(ifname);
+	close(fd);
+
+	if (err < 0)
+		return -EIO;
+
+	memcpy(bssid, wrq.u.ap_addr.sa_data, ETH_ALEN);
+	*bssid_len = ETH_ALEN;
+
+	return 0;
+}
+
 static void add_interface_reply(DBusPendingCall *call, void *user_data)
 {
 	struct supplicant_task *task = user_data;
@@ -269,6 +331,8 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 	return;
 
 failed:
+	dbus_message_unref(reply);
+
 	task_list = g_slist_remove(task_list, task);
 
 	connman_device_unref(task->device);
@@ -290,6 +354,8 @@ static int add_interface(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_message_iter_init_append(message, &array);
 
 	dbus_message_iter_append_basic(&array,
@@ -378,6 +444,8 @@ static int create_interface(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_message_append_args(message, DBUS_TYPE_STRING, &task->ifname,
 							DBUS_TYPE_INVALID);
 
@@ -428,16 +496,20 @@ static int remove_interface(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
+#if 0
 	if (task->created == FALSE) {
 		connman_device_set_powered(task->device, FALSE);
 		return 0;
 	}
+#endif
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, SUPPLICANT_PATH,
 					SUPPLICANT_INTF, "removeInterface");
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &task->path,
 							DBUS_TYPE_INVALID);
 
@@ -475,6 +547,8 @@ static int set_ap_scan(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_message_append_args(message, DBUS_TYPE_UINT32, &ap_scan,
 							DBUS_TYPE_INVALID);
 
@@ -516,6 +590,8 @@ static int add_network(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_error_init(&error);
 
 	reply = dbus_connection_send_with_reply_and_block(connection,
@@ -569,6 +645,8 @@ static int remove_network(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &task->netpath,
 							DBUS_TYPE_INVALID);
 
@@ -611,6 +689,8 @@ static int select_network(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_message_append_args(message, DBUS_TYPE_OBJECT_PATH, &task->netpath,
 							DBUS_TYPE_INVALID);
 
@@ -650,6 +730,8 @@ static int enable_network(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_error_init(&error);
 
 	reply = dbus_connection_send_with_reply_and_block(connection,
@@ -686,6 +768,8 @@ static int disable_network(struct supplicant_task *task)
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_error_init(&error);
 
 	reply = dbus_connection_send_with_reply_and_block(connection,
@@ -715,6 +799,7 @@ static int set_network(struct supplicant_task *task,
 	DBusMessage *message, *reply;
 	DBusMessageIter array, dict;
 	DBusError error;
+	dbus_uint32_t scan_ssid = 1;
 
 	DBG("task %p", task);
 
@@ -726,6 +811,8 @@ static int set_network(struct supplicant_task *task,
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	dbus_message_iter_init_append(message, &array);
 
 	dbus_message_iter_open_container(&array, DBUS_TYPE_ARRAY,
@@ -733,11 +820,10 @@ static int set_network(struct supplicant_task *task,
 			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
-	if (address == NULL) {
-		dbus_uint32_t scan_ssid = 1;
-		connman_dbus_dict_append_variant(&dict, "scan_ssid",
-						DBUS_TYPE_UINT32, &scan_ssid);
-	} else
+	connman_dbus_dict_append_variant(&dict, "scan_ssid",
+					 DBUS_TYPE_UINT32, &scan_ssid);
+
+	if (address)
 		connman_dbus_dict_append_variant(&dict, "bssid",
 						DBUS_TYPE_STRING, &address);
 
@@ -754,14 +840,16 @@ static int set_network(struct supplicant_task *task,
 			connman_dbus_dict_append_variant(&dict, "psk",
 						DBUS_TYPE_STRING, &passphrase);
 	} else if (g_ascii_strcasecmp(security, "wep") == 0) {
-		const char *key_mgmt = "NONE", *index = "0";
+		const char *key_mgmt = "NONE";
 		const char *auth_alg = "OPEN SHARED";
-		connman_dbus_dict_append_variant(&dict, "key_mgmt",
-						DBUS_TYPE_STRING, &key_mgmt);
+		const char *key_index = "0";
 
 		connman_dbus_dict_append_variant(&dict, "auth_alg",
 						DBUS_TYPE_STRING, &auth_alg);
 
+		connman_dbus_dict_append_variant(&dict, "key_mgmt",
+						DBUS_TYPE_STRING, &key_mgmt);
+
 		if (passphrase) {
 			int size = strlen(passphrase);
 			if (size == 10 || size == 26) {
@@ -784,8 +872,9 @@ static int set_network(struct supplicant_task *task,
 				connman_dbus_dict_append_variant(&dict,
 						"wep_key0", DBUS_TYPE_STRING,
 								&passphrase);
+
 			connman_dbus_dict_append_variant(&dict, "wep_tx_keyidx",
-						DBUS_TYPE_STRING, &index);
+						DBUS_TYPE_STRING, &key_index);
 		}
 	} else {
 		const char *key_mgmt = "NONE";
@@ -816,28 +905,66 @@ static int set_network(struct supplicant_task *task,
 	return 0;
 }
 
+static void scan_reply(DBusPendingCall *call, void *user_data)
+{
+	struct supplicant_task *task = user_data;
+	DBusMessage *reply;
+
+	DBG("task %p", task);
+
+	task->scan_call = NULL;
+
+	reply = dbus_pending_call_steal_reply(call);
+	if (reply == NULL)
+		return;
+
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
+		connman_device_set_scanning(task->device, FALSE);
+		goto done;
+	}
+
+	if (task->scanning == TRUE)
+		connman_device_set_scanning(task->device, TRUE);
+
+done:
+	dbus_message_unref(reply);
+}
+
+
 static int initiate_scan(struct supplicant_task *task)
 {
 	DBusMessage *message;
-	DBusPendingCall *call;
 
 	DBG("task %p", task);
 
+	if (task->scan_call != NULL)
+		return -EALREADY;
+
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
 					SUPPLICANT_INTF ".Interface", "scan");
 	if (message == NULL)
 		return -ENOMEM;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	if (dbus_connection_send_with_reply(connection, message,
-						&call, TIMEOUT) == FALSE) {
+					&task->scan_call, TIMEOUT) == FALSE) {
 		connman_error("Failed to initiate scan");
 		dbus_message_unref(message);
 		return -EIO;
 	}
 
+	if (task->scan_call == NULL) {
+		connman_error("D-Bus connection not available");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_pending_call_set_notify(task->scan_call, scan_reply, task, NULL);
+
 	dbus_message_unref(message);
 
-	return 0;
+	return -EINPROGRESS;
 }
 
 static struct {
@@ -876,6 +1003,11 @@ static char *build_group(const char *addr, const char *name,
 	if (str == NULL)
 		return NULL;
 
+	if (ssid == NULL) {
+		g_string_append_printf(str, "hidden_%s", addr);
+		goto done;
+	}
+
 	for (i = 0; special_ssid[i].name; i++) {
 		if (g_strcmp0(special_ssid[i].name, name) == 0) {
 			if (special_ssid[i].value == NULL)
@@ -904,7 +1036,7 @@ static void extract_addr(DBusMessageIter *value,
 					struct supplicant_result *result)
 {
 	DBusMessageIter array;
-	struct ether_addr *eth;
+	struct ether_addr eth;
 	unsigned char *addr;
 	int addr_len;
 
@@ -925,15 +1057,14 @@ static void extract_addr(DBusMessageIter *value,
 	if (result->path == NULL)
 		return;
 
-	eth = (void *) addr;
-
+	memcpy(&eth, addr, sizeof(eth));
 	snprintf(result->path, 13, "%02x%02x%02x%02x%02x%02x",
-						eth->ether_addr_octet[0],
-						eth->ether_addr_octet[1],
-						eth->ether_addr_octet[2],
-						eth->ether_addr_octet[3],
-						eth->ether_addr_octet[4],
-						eth->ether_addr_octet[5]);
+						eth.ether_addr_octet[0],
+						eth.ether_addr_octet[1],
+						eth.ether_addr_octet[2],
+						eth.ether_addr_octet[3],
+						eth.ether_addr_octet[4],
+						eth.ether_addr_octet[5]);
 }
 
 static void extract_ssid(DBusMessageIter *value,
@@ -949,6 +1080,9 @@ static void extract_ssid(DBusMessageIter *value,
 	if (ssid_len < 1)
 		return;
 
+	if (ssid[0] == '\0')
+		return;
+
 	result->ssid = g_try_malloc(ssid_len);
 	if (result->ssid == NULL)
 		return;
@@ -1064,7 +1198,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	unsigned char strength;
 	unsigned short channel, frequency;
 	const char *mode, *security;
-	char *group;
+	char *group = NULL;
+	unsigned int ssid_len;
 
 	DBG("task %p", task);
 
@@ -1202,8 +1337,10 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	if (result.name != NULL && result.name[0] != '\0')
 		connman_network_set_name(network, result.name);
 
-	connman_network_set_blob(network, "WiFi.SSID",
-						result.ssid, result.ssid_len);
+	if (connman_network_get_blob(network, "WiFi.SSID", &ssid_len) == NULL) {
+		connman_network_set_blob(network, "WiFi.SSID",
+					 result.ssid, result.ssid_len);
+	}
 
 	connman_network_set_string(network, "WiFi.Mode", mode);
 
@@ -1218,11 +1355,12 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	connman_network_set_uint16(network, "WiFi.Channel", channel);
 	connman_network_set_string(network, "WiFi.Security", security);
 
-	connman_network_set_group(network, group);
-
-	g_free(group);
+	if (result.ssid != NULL)
+		connman_network_set_group(network, group);
 
 done:
+	g_free(group);
+
 	g_free(result.path);
 	g_free(result.addr);
 	g_free(result.name);
@@ -1236,7 +1374,6 @@ done:
 static void get_properties(struct supplicant_task *task)
 {
 	DBusMessage *message;
-	DBusPendingCall *call;
 	char *path;
 
 	path = g_slist_nth_data(task->scan_results, 0);
@@ -1253,28 +1390,35 @@ static void get_properties(struct supplicant_task *task)
 	if (message == NULL)
 		goto noscan;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	if (dbus_connection_send_with_reply(connection, message,
-						&call, TIMEOUT) == FALSE) {
+				&task->result_call, TIMEOUT) == FALSE) {
 		connman_error("Failed to get network properties");
 		dbus_message_unref(message);
 		goto noscan;
 	}
 
-	if (call == NULL) {
+	if (task->result_call == NULL) {
 		connman_error("D-Bus connection not available");
 		dbus_message_unref(message);
 		goto noscan;
 	}
 
-	dbus_pending_call_set_notify(call, properties_reply, task, NULL);
+	dbus_pending_call_set_notify(task->result_call,
+					properties_reply, task, NULL);
 
 	dbus_message_unref(message);
 
 	return;
 
 noscan:
-	if (task->noscan == FALSE)
+	task->result_call = NULL;
+
+	if (task->scanning == TRUE) {
 		connman_device_set_scanning(task->device, FALSE);
+		task->scanning = FALSE;
+	}
 }
 
 static void scan_results_reply(DBusPendingCall *call, void *user_data)
@@ -1331,38 +1475,47 @@ done:
 	dbus_message_unref(reply);
 
 noscan:
-	if (task->noscan == FALSE)
+	task->result_call = NULL;
+
+	if (task->scanning == TRUE) {
 		connman_device_set_scanning(task->device, FALSE);
+		task->scanning = FALSE;
+	}
 }
 
 static void scan_results_available(struct supplicant_task *task)
 {
 	DBusMessage *message;
-	DBusPendingCall *call;
 
 	DBG("task %p", task);
 
+	if (task->result_call != NULL)
+		return;
+
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
 						SUPPLICANT_INTF ".Interface",
 							"scanResults");
 	if (message == NULL)
 		return;
 
+	dbus_message_set_auto_start(message, FALSE);
+
 	if (dbus_connection_send_with_reply(connection, message,
-						&call, TIMEOUT) == FALSE) {
+				&task->result_call, TIMEOUT) == FALSE) {
 		connman_error("Failed to request scan result");
 		goto done;
 	}
 
-	if (task->noscan == FALSE)
-		connman_device_set_scanning(task->device, TRUE);
-
-	if (call == NULL) {
+	if (task->result_call == NULL) {
 		connman_error("D-Bus connection not available");
 		goto done;
 	}
 
-	dbus_pending_call_set_notify(call, scan_results_reply, task, NULL);
+	if (task->scanning == TRUE)
+		connman_device_set_scanning(task->device, TRUE);
+
+	dbus_pending_call_set_notify(task->result_call,
+					scan_results_reply, task, NULL);
 
 done:
 	dbus_message_unref(message);
@@ -1395,6 +1548,7 @@ static int task_connect(struct supplicant_task *task)
 	const char *address, *security, *passphrase;
 	const void *ssid;
 	unsigned int ssid_len;
+	int err;
 
 	address = connman_network_get_string(task->network, "Address");
 	security = connman_network_get_string(task->network, "WiFi.Security");
@@ -1418,15 +1572,19 @@ static int task_connect(struct supplicant_task *task)
 
 	set_network(task, ssid, ssid_len, address, security, passphrase);
 
-	enable_network(task);
+	err = enable_network(task);
+	if (err < 0)
+		return err;
 
-	return 0;
+	return -EINPROGRESS;
 }
 
 static void state_change(struct supplicant_task *task, DBusMessage *msg)
 {
 	DBusError error;
 	const char *newstate, *oldstate;
+	unsigned char bssid[ETH_ALEN];
+	unsigned int bssid_len;
 	enum supplicant_state state;
 
 	dbus_error_init(&error);
@@ -1448,43 +1606,33 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 	if (state == WPA_INVALID)
 		return;
 
-	task->state = state;
-
-	switch (task->state) {
-	case WPA_SCANNING:
-		task->noscan = TRUE;
-		connman_device_set_scanning(task->device, TRUE);
-		break;
-	case WPA_ASSOCIATING:
-	case WPA_ASSOCIATED:
-	case WPA_4WAY_HANDSHAKE:
-	case WPA_GROUP_HANDSHAKE:
-		task->noscan = TRUE;
-		break;
-	case WPA_COMPLETED:
-	case WPA_DISCONNECTED:
-		task->noscan = FALSE;
-		break;
-	case WPA_INACTIVE:
-		task->noscan = FALSE;
+	if (task->scanning == TRUE && state != WPA_SCANNING) {
 		connman_device_set_scanning(task->device, FALSE);
-		break;
-	case WPA_INVALID:
-		break;
+		task->scanning = FALSE;
 	}
 
+	task->state = state;
+
 	if (task->network == NULL)
 		return;
 
 	switch (task->state) {
 	case WPA_COMPLETED:
+		if (get_bssid(task->device, bssid, &bssid_len) == 0)
+	                connman_network_set_address(task->network,
+							bssid, bssid_len);
+
 		/* carrier on */
 		connman_network_set_connected(task->network, TRUE);
-		connman_device_set_scanning(task->device, FALSE);
 		break;
+
 	case WPA_DISCONNECTED:
+		disable_network(task);
+
+		/* carrier off */
+		connman_network_set_connected(task->network, FALSE);
+
 		if (task->disconnecting == TRUE) {
-			connman_network_set_connected(task->network, FALSE);
 			connman_network_unref(task->network);
 			task->disconnecting = FALSE;
 
@@ -1492,16 +1640,31 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 				task->network = task->pending_network;
 				task->pending_network = NULL;
 				task_connect(task);
-			}
-		} else {
-			/* carrier off */
-			connman_network_set_connected(task->network, FALSE);
-			connman_device_set_scanning(task->device, FALSE);
+			} else
+				task->network = NULL;
 		}
 		break;
+
 	case WPA_ASSOCIATING:
 		connman_network_set_associating(task->network, TRUE);
 		break;
+
+	case WPA_INACTIVE:
+		connman_network_set_connected(task->network, FALSE);
+
+		if (task->disconnecting == TRUE) {
+			connman_network_unref(task->network);
+			task->disconnecting = FALSE;
+
+			if (task->pending_network != NULL) {
+				task->network = task->pending_network;
+				task->pending_network = NULL;
+				task_connect(task);
+			} else
+				task->network = NULL;
+		}
+		break;
+
 	default:
 		connman_network_set_associating(task->network, FALSE);
 		break;
@@ -1540,27 +1703,6 @@ static DBusHandlerResult supplicant_filter(DBusConnection *conn,
 	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
-static int supplicant_get_range(struct supplicant_task *task)
-{
-	struct iwreq wrq;
-	int fd, err;
-
-	fd = socket(PF_INET, SOCK_DGRAM, 0);
-	if (fd < 0)
-		return -1;
-
-	memset(&wrq, 0, sizeof(struct iwreq));
-	strncpy(wrq.ifr_name, task->ifname, IFNAMSIZ);
-	wrq.u.data.pointer = task->range;
-	wrq.u.data.length = sizeof(struct iw_range);
-
-	err = ioctl(fd, SIOCGIWRANGE, &wrq);
-
-	close(fd);
-
-	return err;
-}
-
 int supplicant_start(struct connman_device *device)
 {
 	struct supplicant_task *task;
@@ -1586,14 +1728,14 @@ int supplicant_start(struct connman_device *device)
 		goto failed;
 	}
 
-	err = supplicant_get_range(task);
+	err = get_range(task);
 	if (err < 0)
 		goto failed;
 
 	task->device = connman_device_ref(device);
 
 	task->created = FALSE;
-	task->noscan = FALSE;
+	task->scanning = FALSE;
 	task->state = WPA_INVALID;
 	task->disconnecting = FALSE;
 	task->pending_network = NULL;
@@ -1625,6 +1767,19 @@ int supplicant_stop(struct connman_device *device)
 
 	task_list = g_slist_remove(task_list, task);
 
+	if (task->scan_call != NULL) {
+		dbus_pending_call_cancel(task->scan_call);
+		task->scan_call = NULL;
+	}
+
+	if (task->result_call != NULL) {
+		dbus_pending_call_cancel(task->result_call);
+		task->result_call = NULL;
+	}
+
+	if (task->scanning == TRUE)
+		connman_device_set_scanning(task->device, FALSE);
+
 	disable_network(task);
 
 	remove_network(task);
@@ -1656,7 +1811,18 @@ int supplicant_scan(struct connman_device *device)
 		break;
 	}
 
+	task->scanning = TRUE;
+
 	err = initiate_scan(task);
+	if (err < 0) {
+		if (err == -EINPROGRESS)
+			return 0;
+
+		task->scanning = FALSE;
+		return err;
+	}
+
+	connman_device_set_scanning(task->device, TRUE);
 
 	return 0;
 }
@@ -1681,7 +1847,7 @@ int supplicant_connect(struct connman_network *network)
 		return task_connect(task);
 	}
 
-	return 0;
+	return -EINPROGRESS;
 }
 
 int supplicant_disconnect(struct connman_network *network)
@@ -1698,7 +1864,7 @@ int supplicant_disconnect(struct connman_network *network)
 		return -ENODEV;
 
 	if (task->disconnecting == TRUE)
-		return -EINPROGRESS;
+		return -EALREADY;
 
 	disable_network(task);
 
