@@ -255,16 +255,11 @@ static int set_connected(struct connman_device *device,
 
 static int set_carrier(struct connman_device *device, connman_bool_t carrier)
 {
-	struct connman_service *service;
-
 	if (carrier == TRUE)
 		__connman_profile_add_device(device);
 	else
 		__connman_profile_remove_device(device);
 
-	service = __connman_service_lookup_from_device(device);
-	__connman_service_set_carrier(service, carrier);
-
 	return set_connected(device, carrier);
 }
 
@@ -676,6 +671,7 @@ static void unregister_interface(struct connman_element *element)
 
 static int setup_device(struct connman_device *device)
 {
+	enum connman_service_type type;
 	int err;
 
 	DBG("device %p", device);
@@ -688,6 +684,9 @@ static int setup_device(struct connman_device *device)
 		return err;
 	}
 
+	type = __connman_device_get_service_type(device);
+	__connman_notifier_register(type);
+
 	switch (device->mode) {
 	case CONNMAN_DEVICE_MODE_UNKNOWN:
 	case CONNMAN_DEVICE_MODE_NETWORK_SINGLE:
@@ -731,6 +730,8 @@ static void probe_driver(struct connman_element *element, gpointer user_data)
 
 static void remove_device(struct connman_device *device)
 {
+	enum connman_service_type type;
+
 	DBG("device %p", device);
 
 	__connman_device_disable(device);
@@ -746,6 +747,9 @@ static void remove_device(struct connman_device *device)
 		break;
 	}
 
+	type = __connman_device_get_service_type(device);
+	__connman_notifier_unregister(type);
+
 	unregister_interface(&device->element);
 
 	if (device->driver->remove)
@@ -1629,22 +1633,12 @@ int __connman_device_set_offlinemode(connman_bool_t offlinemode)
 
 void __connman_device_increase_connections(struct connman_device *device)
 {
-	enum connman_service_type type;
-
 	device->connections++;
-
-	type = __connman_device_get_service_type(device);
-	__connman_notifier_connect(type);
 }
 
 void __connman_device_decrease_connections(struct connman_device *device)
 {
-	enum connman_service_type type;
-
 	device->connections--;
-
-	type = __connman_device_get_service_type(device);
-	__connman_notifier_disconnect(type);
 }
 
 /**
@@ -1752,15 +1746,10 @@ void __connman_device_set_network(struct connman_device *device,
  */
 int connman_device_register(struct connman_device *device)
 {
-	enum connman_service_type type;
-
 	__connman_storage_load_device(device);
 
 	device->offlinemode = __connman_manager_get_offlinemode();
 
-	type = __connman_device_get_service_type(device);
-	__connman_notifier_register(type);
-
 	return connman_element_register(&device->element, NULL);
 }
 
@@ -1772,13 +1761,8 @@ int connman_device_register(struct connman_device *device)
  */
 void connman_device_unregister(struct connman_device *device)
 {
-	enum connman_service_type type;
-
 	__connman_storage_save_device(device);
 
-	type = __connman_device_get_service_type(device);
-	__connman_notifier_unregister(type);
-
 	connman_element_unregister(&device->element);
 }
 
